/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 2002 The Apache Software Foundation.  All rights 
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:  
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "WSIF" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written 
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 2001, 2002, International
 * Business Machines, Inc., http://www.apache.org.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

package org.apache.wsif.providers.ejb;

import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;

import javax.ejb.EJBHome;
import javax.ejb.EJBObject;
import javax.naming.InitialContext;
import javax.rmi.PortableRemoteObject;
import javax.wsdl.BindingOperation;
import javax.wsdl.Definition;
import javax.wsdl.Port;
import javax.wsdl.extensions.ExtensibilityElement;

import org.apache.wsif.WSIFException;
import org.apache.wsif.WSIFOperation;
import org.apache.wsif.logging.Trc;
import org.apache.wsif.providers.WSIFDynamicTypeMap;
import org.apache.wsif.util.WSIFUtils;
import org.apache.wsif.wsdl.extensions.ejb.EJBAddress;

/**
 * EJB WSIF Port.
 * @author <a href="mailto:gpfau@de.ibm.com">Gerhard Pfau</a>
 * Partially based on WSIFPort_ApacheSOAP from Alekander Slominski,
 * Paul Fremantle, Sanjiva Weerawarana and Matthew J. Duftler
 * @author Owen Burroughs <owenb@apache.org>
 * @author Jeremy Hughes <hughesj@apache.org>
 */
public class WSIFPort_EJB
    extends org.apache.wsif.base.WSIFDefaultPort
    implements java.io.Serializable {

	private static final long serialVersionUID = 1L;

    private javax.wsdl.Definition fieldDefinition = null;
    private javax.wsdl.Port fieldPortModel = null;
    private EJBHome fieldEjbHome = null; // 'factory for physical connection'
    private EJBObject fieldEjbObject = null; // 'physical connection'

    protected Map operationInstances = new HashMap();

    public WSIFPort_EJB(Definition def, Port port, WSIFDynamicTypeMap typeMap) {
        Trc.entry(this, def, port, typeMap);

        fieldDefinition = def;
        fieldPortModel = port;

        Trc.exit();
    }

    public Definition getDefinition() {
        return fieldDefinition;
    }

    public WSIFOperation_EJB getDynamicWSIFOperation(
        String name,
        String inputName,
        String outputName)
        throws WSIFException {
        Trc.entry(this, name, inputName, outputName);

        WSIFOperation_EJB tempOp =
            (WSIFOperation_EJB) operationInstances.get(getKey(name, inputName, outputName));

		WSIFOperation_EJB operation = null;
		if (tempOp != null) {
			operation = tempOp.copy();
		}

        if (operation == null) {
            BindingOperation bindingOperationModel =
               WSIFUtils.getBindingOperation( 
                  fieldPortModel.getBinding(), name, inputName, outputName );

            if (bindingOperationModel != null) {
                operation = new WSIFOperation_EJB(fieldPortModel, bindingOperationModel, this);
                setDynamicWSIFOperation(name, inputName, outputName, operation);
            }
        }

        Trc.exit(operation);
        return operation;
    }

    public EJBHome getEjbHome() throws WSIFException {
        Trc.entry(this);

        if (fieldEjbHome == null) {
            EJBAddress address = null;

            try {
                ExtensibilityElement portExtension =
                    (ExtensibilityElement) fieldPortModel.getExtensibilityElements().get(0);

                if (portExtension == null) {
                    throw new WSIFException("missing port extension");
                }

                address = (EJBAddress) portExtension;

                Hashtable hash = new Hashtable();

                // If the initial context factory is specified in the wsdl use it
                String icf = address.getInitialContextFactory();
                if (icf != null) {
                    hash.put(InitialContext.INITIAL_CONTEXT_FACTORY, icf);
                }

                // If the jndi provider url is specified in the wsdl use it
                String providerURL = address.getJndiProviderURL();
                if (providerURL != null) {
                    hash.put(InitialContext.PROVIDER_URL, providerURL);
                }

                // Lookup from an authoritative source								
                hash.put(InitialContext.AUTHORITATIVE, "true");

                InitialContext initContext;
                initContext = new InitialContext(hash);
                
                Class homeClass = null;
                try {
                    if (address.getClassName() != null) {
                        homeClass =
                            Class.forName(
                                address.getClassName(),
                                true,
                                Thread.currentThread().getContextClassLoader());
                        if (!(EJBHome.class.isAssignableFrom(homeClass))) {
                            homeClass = null;
                        }
                    }
                } catch (ClassNotFoundException cnf) {
                	Trc.ignoredException(cnf);
                }
                
                if (homeClass != null) {
                	fieldEjbHome = (EJBHome) PortableRemoteObject.narrow(initContext.lookup(address.getJndiName()),homeClass);
                } else {
                	fieldEjbHome = (EJBHome) initContext.lookup(address.getJndiName());
                }

            } catch (Exception ex) {
            	Trc.exception(ex);
                throw new WSIFException(
                    "Failed to lookup EJB home using JNDI name '" + address.getJndiName() + "'",
                    ex);
            }
        }
        Trc.exit(fieldEjbHome);
        return fieldEjbHome;
    }

    public EJBObject getEjbObject() throws WSIFException {
        Trc.entry(this);

        if (fieldEjbObject == null) {
            EJBHome ejbHome = getEjbHome();
            try {
                Method createMethod = ejbHome.getClass().getDeclaredMethod("create", null);
                fieldEjbObject = (EJBObject) createMethod.invoke(ejbHome, null);
            } catch (Exception ex) {
            	Trc.exception(ex);
                throw new WSIFException(
                    "Could not create instance for home '" + ejbHome + "'",
                    ex);
            }
        }
        Trc.exit(fieldEjbObject);
        return fieldEjbObject;
    }

    public Port getPortModel() {
        Trc.entry(this);
        Trc.exit(fieldPortModel);
        return fieldPortModel;
    }

    public void setDefinition(Definition value) {
        Trc.entry(this, value);
        fieldDefinition = value;
        Trc.exit();
    }

    // WSIF: keep list of operations available in this port
    public void setDynamicWSIFOperation(
        String name,
        String inputName,
        String outputName,
        WSIFOperation_EJB value) {
        Trc.entry(this, name, inputName, outputName);

        operationInstances.put(getKey(name, inputName, outputName), value);
        Trc.exit();
    }

    public void setEjbHome(EJBHome newEjbHome) {
        Trc.entry(this, newEjbHome);
        fieldEjbHome = newEjbHome;
        Trc.exit();
    }

    public void setEjbObject(EJBObject newEjbObject) {
        Trc.entry(this, newEjbObject);
        fieldEjbObject = newEjbObject;
        Trc.exit();
    }

    public void setPortModel(Port value) {
        Trc.entry(this, value);
        fieldPortModel = value;
        Trc.exit();
    }

    public WSIFOperation createOperation(String operationName)
        throws WSIFException {
        Trc.entry(this, operationName);
        WSIFOperation wo = createOperation(operationName, null, null);
        Trc.exit(wo);
        return wo;
    }

    public WSIFOperation createOperation(
        String operationName,
        String inputName,
        String outputName)
        throws WSIFException {
        Trc.entry(this, operationName, inputName, outputName);
        WSIFOperation_EJB op =
            getDynamicWSIFOperation(operationName, inputName, outputName);
        if (op == null) {
            throw new WSIFException(
                "Could not create operation: "
                    + operationName
                    + ":"
                    + inputName
                    + ":"
                    + outputName);
        }
        WSIFOperation wo = op.copy();
        Trc.exit(wo);
        return wo;
    }

    public String deep() {
        String buff = "";
        try {
            buff = new String(super.toString() + ":\n");

            buff += "definition:" + Trc.brief(fieldDefinition);
            buff += " portModel:" + Trc.brief(fieldPortModel);
            buff += " ejbHome:" + fieldEjbHome;
            buff += " ejbObject" + fieldEjbObject;

            buff += " operationInstances:";
            if (operationInstances == null)
                buff += "null";
            else {
                buff += " size:" + operationInstances.size();
                Iterator it = operationInstances.keySet().iterator();
                int i = 0;
                while (it.hasNext()) {
                    String key = (String) it.next();
                    WSIFOperation_EJB woejb = (WSIFOperation_EJB) operationInstances.get(key);
                    buff += "\noperationInstances[" + i + "]:" + key + " " + woejb + " ";
                    i++;
                }
            }
        } catch (Exception e) {
            Trc.exceptionInTrace(e);
        }

        return buff;
    }
}