/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 2002 The Apache Software Foundation.  All rights 
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:  
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "WSIF" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written 
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 2001, 2002, International
 * Business Machines, Inc., http://www.apache.org.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

package soap;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import org.apache.wsif.WSIFException;
import org.apache.wsif.WSIFMessage;
import org.apache.wsif.WSIFOperation;
import org.apache.wsif.WSIFPort;
import org.apache.wsif.WSIFService;
import org.apache.wsif.WSIFServiceFactory;
import org.apache.wsif.providers.soap.apacheaxis.WSIFDynamicProvider_ApacheAxis;
import org.apache.wsif.util.WSIFPluggableProviders;
import util.TestUtilities;

import addressbook.wsiftypes.Address;
import addressbook.wsiftypes.Phone;

/**
 * Junit test to test out input message parts using the AddressBook sample.
 * 
 * The WSDL input message can define multiple parts, but the parts in the 
 * WSIFMessage passed to the WSIFOperation may not match what the WSDL defines. 
 * Its been decided the parts defined in the WSDL but not in the WSIFMessage 
 * should be sent as null, and parts in the WSIFMessage but not in the WSDL 
 * should be ignored.
 *   
 * @author Ant Elder <ant.elder@uk.ibm.com>
 */
public class InputPartsTest extends TestCase {
    static final String WSDL_LOCATION =
        TestUtilities.getWsdlPath("java\\test\\addressbook\\wsifservice") +
        "AddressBook.wsdl";
    static String server = TestUtilities.getSoapServer().toUpperCase();

    public InputPartsTest(String name) {
        super(name);
    }

    public static void main(String[] args)
    {
        TestUtilities.startListeners(
            TestUtilities.ADDRESSBOOK_LISTENER
                | TestUtilities.NATIVEJMS_LISTENER);

        junit.textui.TestRunner.run(suite());
        TestUtilities.stopListeners();
    }

    public static Test suite() {
        return new TestSuite(InputPartsTest.class);
    }

    public void setUp() {
        TestUtilities.setUpExtensionsAndProviders();
    }

    public void testSoap() {
        doit(server+"Port", "soap");
    }
    public void testJava() {
        doit("JavaPort", "java");
    }
    public void testSoapJms() {
        doit("SOAPJMSPort", "soap");
    }
//TODO the SOAP service doesn't work with AXIS nulls
//    public void testAxis() {
//        doit(server+"Port", "axis");
//    }
//    public void testAxisJms() {
//        doit("SOAPJMSPort", "axis");
//    }
    public void testNativeJms() {
        doit("NativeJmsPort", "" ); 
    }

    private void doit(String portName, String protocol) {
        if (portName.toUpperCase().indexOf("JMS") != -1 && !TestUtilities.areWeTesting("jms"))
            return;

        TestUtilities.setProviderForProtocol( protocol );

        System.out.println(
           "\n=== " + this.getClass().getName() + " " + portName + " " + protocol );

        try {
            WSIFServiceFactory factory = WSIFServiceFactory.newInstance();
                WSIFService service = factory.getService(WSDL_LOCATION,
                null, // serviceNS 
                null, // serviceName 
                "http://wsifservice.addressbook/", // portTypeNS 
                "AddressBook"); // portTypeName 

            service.mapType(
                new javax.xml.namespace.QName(
                    "http://wsiftypes.addressbook/",
                    "address"),
                Class.forName("addressbook.wsiftypes.Address"));

            service.mapType(
                new javax.xml.namespace.QName(
                    "http://wsiftypes.addressbook/",
                    "phone"),
                Class.forName("addressbook.wsiftypes.Phone"));

            WSIFPort port = service.getPort( portName );

            Address addr1 = new Address (1, "University Drive",
                                      "West Lafayette", "IN", 47907,
                                      new Phone (765, "494", "4900"));
            Address addr2 = new Address (0, "Somewhere Else",
                                      "No Where", "NO", 71983,
                                      new Phone (600, "391", "5682"));
            Address addr3 = new Address( 0,"","","",0, new Phone() );

            addName( port, "ant", addr1 );
            Address addressFound = getAddress( port, "ant" );
            assertTrue( "test 1 addresses are not equal!!", addressFound.equals( addr1 ) );

            addName( port, null, addr2 );
            addressFound = getAddress( port, null );
            assertTrue( "test 2 addresses are not equal!!", addressFound.equals( addr2 ) );

            addName( port, "petra", null );
            addressFound = getAddress( port, "petra" );
//TODO Address gets NPE on null address compare 
//          assertTrue( "test 3 addresses are not equal!!", addressFound.equals( addr3 ) );

        } catch (Exception e) {
            e.printStackTrace();
            assertTrue( "got exception: " + e, false);
        } finally {
            if (protocol.equals("axis")) {
			   WSIFPluggableProviders.overrideDefaultProvider(
                    "http://schemas.xmlsoap.org/wsdl/soap/",
                    null);
            }
        }

    }

    private void addName(WSIFPort port, String name, Address addr) 
       throws WSIFException {
       	
       WSIFOperation operation = 
          port.createOperation("addEntry", "AddEntryWholeNameRequest", null); 
       WSIFMessage inputMessage = operation.createInputMessage();
       WSIFMessage outputMessage = operation.createOutputMessage();
       WSIFMessage faultMessage = operation.createFaultMessage();

       if ( name != null ) {
          inputMessage.setObjectPart( "name", name );
       }
       if ( addr != null ) {
          inputMessage.setObjectPart( "address", addr );
       }
       inputMessage.setObjectPart( "extra", "junk" );

       boolean ok = operation.executeRequestResponseOperation(
          inputMessage,
          outputMessage,
          faultMessage);

       assertTrue( "addEntry operation returned false!!", ok );
       
    }
    
    private Address getAddress(WSIFPort port, String name) 
       throws WSIFException {

       WSIFOperation operation = port.createOperation("getAddressFromName");
       WSIFMessage inputMessage = operation.createInputMessage();
       WSIFMessage outputMessage = operation.createOutputMessage();
       WSIFMessage faultMessage = operation.createFaultMessage();

       if ( name != null ) {
          inputMessage.setObjectPart( "name", name );
       }
       inputMessage.setObjectPart( "extra", "junk" );

       boolean ok = operation.executeRequestResponseOperation(
          inputMessage,
          outputMessage,
          faultMessage);

       assertTrue( "getAddressFromName operation returned false!!", ok );
       
       return (Address) outputMessage.getObjectPart( "address" ); 

    }
    
}
